# superclass
require 'khttpsv_template.rb'

# file extension and mime type mapping hash
require 'mime_type.rb'

# getpwent
require 'etc'
# open3, bi-directional pipe
require 'open3'

class Khttpsv_simple < Khttpsv_template
  #----------------------------------------------------------------------
  # HTTP/1.0 server that implement basic file transfer
  #----------------------------------------------------------------------
  # INSTANCE VARIABLES
  # @translated_path: String
  # @index_file: String
  # @user_document_dir: String
  #----------------------------------------------------------------------
  # METHODS
  # def allow?(a)
  # def cgi?
  # def document_root
  # def translate_path
  # def hook_translate_path(s)
  # def get_mime_type
  # def dir_handler
  # def cgi_handler
  # def file_handler

  #----------------------------------------------------------------------
  # feature policy manager
  def allow?(a)
    case a
    when 'user_dir'
      true
    when 'dir_indexing'
      true
    else
      # 'cgi_exec', 'ssi_exec' and other future features
      false
    end
  end

  #----------------------------------------------------------------------
  #
  def cgi?
    @translated_path[-4,4].downcase == '.cgi' && FileTest.executable?(@translated_path)
  end

  #----------------------------------------------------------------------
  # determine document root --
  # virtual server may override this method (such method will
  # return different value according to @req_header['Host'])
  def document_root
    '/usr/local/www/htdocs'
  end

  #----------------------------------------------------------------------
  # convert URI to absolute real path
  def translate_path
    if @uri_encoded[0] != ?/ then
      @uri_encoded = '/' + @uri_encoded
    end

    begin
      if tmp = hook_translate_path(@uri_encoded) then
	@translated_path = tmp
      elsif allow?('user_dir') && @uri_encoded =~ %r{^/~([^/]+)/} then
	# http://server/~user/
	# getpwnam may raise exception for non-existent user
	pw = Etc.getpwnam($1)
	@translated_path = pw.dir + '/' + @user_document_dir + '/' + $'
      else
	raise ArgumentError
      end
    rescue ArgumentError
      droot = document_root()
      @translated_path = droot + @uri_encoded
    end

    @translated_path
  end

  #----------------------------------------------------------------------
  # hook to translation of uri
  def hook_translate_path(s) return nil end

  #----------------------------------------------------------------------
  # determine MIME content-type by its extension
  def get_mime_type
    ctype = 'text/plain'
    if tmp = @translated_path.rindex('.') then
      ext = @translated_path[tmp + 1, @translated_path.length - tmp - 1]
      ext.downcase!
      ctype = MIME_map[ext] if MIME_map[ext]
    end
    return ctype
  end

  #----------------------------------------------------------------------
  def get_server_addr
    # IP address of incoming connection
    # overriding this method that returns DNS name is prefferable
    @sock.addr[3]
  end

  #----------------------------------------------------------------------
  #
  def dir_handler
    @resp_headers['content-type'] = ['text/html']

    if @translated_path[-1] != ?/ then
      moved_to = sprintf('http://%s:%d%s/', @sock.addr[2], @sock.addr[1], @uri_raw)
      @resp_headers['location'] = [moved_to]

      send_status_header(STAT_MOVED_TEMP, "Moved temporarily")
      send_response_header
      begin_content
      send_content("Resource <a href=\"#{@uri_raw}\">#{@uri_raw}</a> you have requested has moved <a href=\"#{moved_to}/\">here</a>")
      return
    end

    begin
      tmp = File.expand_path(@index_file, @translated_path)
      if FileTest.file?(tmp) then
	# treat "http://server/foo/" as if "http://server/foo/index.html"
	@translated_path = tmp
	return file_handler()
      end

      if !allow?('dir_indexing') then
	logger('attempt to list directory', LOG_WARN)
	send_status_header(STAT_FORBIDDEN, "Directory listing aborted")
	send_response_header
	return
      end

      dir = Dir.open(@translated_path)
      send_status_header(STAT_OK, "Ok")
      send_response_header
      begin_content
      send_content("<html><body>\n")
      send_content("<h1>Index of " + @uri_raw + "</h1>\n")
      send_content("<pre>")

      dir.each do |file|
	next if file == '.' || file == '..'
	tmp = File.expand_path(file, @translated_path)
	mtime = File.stat(tmp).mtime
	send_content(sprintf("<a href=\"%s\">%-40s</a> %-30s\n",
			     url_escape(file),
			     html_escape(file),
			     mtime.to_s))
      end
      send_content("</pre></body></html>\n")
      dir.close
    rescue Errno::ENOENT
      logger('attempt to get non-existent content', LOG_WARN)
      send_status_header(STAT_NOT_FOUND, "Not found")
      send_response_header
    end
  end

  #----------------------------------------------------------------------
  # 
  def cgi_handler
    # check if CGI invocation is allowed?
    if !allow?('cgi_exec')
      logger('attempt to exec CGI that config inhibits', LOG_WARN)
      send_status_header(STAT_NOT_IMPLEMENTED, 'CGI not implemented')
      send_response_header
      begin_content
      return
    end

    # build environment variables
    ENV.each do |k, v|
      ENV.delete(k)
    end
    ENV['PATH'] = '/bin:/usr/bin'
    ENV['SERVER_SOFTWARE'] = @server_name
    ENV['SERVER_NAME'] = @server_addr
    ENV['GATEWAY_INTERFACE'] = 'CGI/1.0'
    ENV['SERVER_PROTOCOL'] = 'HTTP/1.0'
    ENV['SERVER_PORT'] = @sock.addr[1].to_s
    ENV['REQUEST_METHOD'] = @http_method
    ENV['SCRIPT_NAME'] = @uri_raw
    if(@sock.peeraddr[2] != @sock.peeraddr[3])
      # has DNS name
      ENV['REMOTE_HOST'] = @sock.peeraddr[2]
    end
    ENV['REMOTE_ADDR'] = @sock.peeraddr[3]
    if @req_headers['Content-type'] then
      ENV['CONTENT_TYPE'] = @req_headers['Content-type'][0]
    end
    if @http_method == 'POST' then
      if @req_headers['Content-length'] then
	post_size = @req_headers['Content-length'][0]
      else
	# fail safe for server attack
	post_size = '0'
      end
      ENV['CONTENT_LENGTH'] = post_size
    end
    # TODO: PATH_INFO,PATH_TRANSLATED,AUTH_TYPE,REMOTE_USER,REMOTE_IDENT
    if @http_method == 'GET' then
      ENV['QUERY_STRING'] = @uri_query
    end
    @req_headers.each do |key, val|
      ENV['HTTP_' + key.upcase] = val.join(', ')
    end

    begin
      null = open('/dev/null')
      pin, pout, perr = Open3.popen3(@translated_path)
      perr.reopen(null)
      null.close

      if @http_method == 'POST' then
	# write post data to CGI script
	remain = post_size.to_i
	th = Thread.new do
	  while(remain > 0 && buf = @sock.read(remain)) do
	    logger("cgi_post: read #{buf.size} bytes from client, #{buf.dump}", LOG_DEBUG)
	    pin.write(buf)
	    remain -= buf.size
	  end
	end
      end

      # if not nph script, read directives from CGI
      if File.basename(@translated_path) != 'nph-' then
	status = '200 Ok'
	location = nil
	ctype = 'text/plain'
	while s = pout.gets.chop do
	  if s == '' then
	    break
	  elsif s =~ /^status: (^\d{3} .*)/i then
	    status = $1 
	  elsif s =~ /^location: (.*)/i then
	    location = $1
	  elsif s =~ /^content-type: (.*)/i then
	    ctype = $1
	  elsif s =~ /^(.*): (.*)$/ then
	    key = $1.downcase
	    if @resp_headers[key].is_a?(Array) then
	      @resp_headers[key] << $2
	    else
	      @resp_headers[key] = [$2]
	    end
	  else
	    logger('err malformed directive from CGI: ' + s.dump, LOG_ERR)
	  end
	end

	# send client to HTTP headers
	@resp_headers['content-type'] = [ctype]
	@resp_headers['location'] = [location] if location
	send_status_header(status, "")
	send_response_header
	begin_content
      end

      # read from CGI and write it to client
      while s = pout.gets do
	send_content(s)
      end
    rescue
      logger('err cgi exec error: ' + $!.inspect + ' at ' + $@.join(' > '), LOG_ERR)
    ensure
      pin.close
      pout.close
      perr.close
      if @http_method == 'POST' then
	th.join
      end
    end
  end

  #----------------------------------------------------------------------
  # 
  def file_handler
    # reject unsupported method
    unless @http_method == 'GET' || @http_method == 'HEAD'
      logger('unsupported method requested', LOG_ERR)
      send_status_header(STAT_NOT_IMPLEMENTED, 'Unsupported method')
      return
    end

    file = @translated_path
    begin
      @resp_headers['content-type'] = ['text/html']

      st = File.stat(file)
      fd = open(file, 'r')
      @resp_headers['last-modified'] = [rfc1123date(st.mtime)]
      @resp_headers['content-length'] = [st.size.to_s]

      # now here, file exists and is readable, so
      # we can determine its content
      ctype = get_mime_type()
      @resp_headers['content-type'] = [ctype]

      # If-modified-since header
      # TODO: "length" option
      if @req_headers['If-modified-since'] then
	tmp = @req_headers['If-modified-since'][0].sub(/;.*/, '')
	ims = parse_rfc1123date(tmp)
	if ims && st.mtime - ims <= 0 && ims - Time.now < 0 then
	  send_status_header(STAT_NOT_MODIFIED, "Not modified")
	  send_response_header
	  begin_content
	  fd.close
	  raise # pseudo error
	end
      end

      begin
	send_status_header(STAT_OK, "Ok")
	send_response_header
	begin_content
	while TRUE
	  # may occur EOFError and Errno::EPIPE
	  buf = fd.sysread(16 * 1024)
	  send_content(buf)
	end
      rescue EOFError
      rescue Errno::EPIPE
	# nothing special to do
      ensure
	fd.close
      end

    rescue Errno::ENOENT
      # failed to stat() or open()
      logger(@translated_path.dump + ' not found', LOG_ERR)
      send_status_header(STAT_NOT_FOUND, "Not found")
      send_response_header
      begin_content
      send_content("Resource <a href=\"#{@uri_raw}\">#{@uri_raw}</a> you have requested is not found on this server")
    rescue
      # content transfer omitted due to ims header
    end
    #send_content("Resource <a href=\"#{@uri_raw}\">#{@uri_raw}</a> you have requested is forbidden to serve")
  end

  #----------------------------------------------------------------------
  def dispatch_handler
    @server_addr = get_server_addr()
    translate_path()
    logger("path " + @translated_path.dump, LOG_INFO)

    # TODO: access restriction here

    if cgi?() then
      # memo: dont apply file_handler() for any ".cgi" scripts
      # even if CGI invocation is inhibited.
      cgi_handler()
    elsif FileTest.directory?(@translated_path) then
      dir_handler()
    else
      file_handler()
    end
  end

  #----------------------------------------------------------------------
  #----------------------------------------------------------------------
  # 
  def initialize(s)
    super
    @server_name = 'SimpleRubyHTTPServer/1.0.0' # override
    @index_file = 'index.html'
    @user_document_dir = '.public_html'
  end

  #----------------------------------------------------------------------
  # 
  def Khttpsv_simple.start(req_port = 0)
    generic_sock = TCPServer.open(req_port)
    opened_port = generic_sock.addr[1]
    logger("waiting at port " + opened_port.to_s, LOG_NOTICE)

    begin
      while TRUE
	ns = generic_sock.accept
	Thread.start do
	  ns2 = ns # save to seperate stack
	  server = new(ns2)
	  server.do_work
	  ns2.close
	end
      end

    rescue
      logger('err unhandled error: ' + $!.inspect + ' at ' + $@.join(' > '), LOG_ERR)
    end
  end

  #----------------------------------------------------------------------
  # 
  trap('CHLD') do
    begin
      while pid = Process.waitpid(-1, Process::WNOHANG) do
	logger("cgi_exit pid #{pid}", LOG_INFO)
      end
    rescue Errno::ECHILD
      # no child to wait
    end
  end
end
