#!/usr/local/bin/perl

# Reference:
#  [1] "Hypertext Transfer Protocol HTTP/1.0", RFC1945
#  [2] "$B$R$m$N$N(BCGI tips", http://www.st.rim.or.jp/~hirono/cgitips/tips.html
#  [3] "The CGI Specification", http://hoohoo.ncsa.uiuc.edu/cgi/env.html

$conf_server = "test perl server/1.0";
$conf_defdoc = "index.html";
$conf_deftype = "text/html";
$conf_usrdir = ".public_html";
$conf_docroot = "/usr/local/etc/httpd/htdocs";
%conf_exttype = ("html", "text/html",
		 "doc", "text/plain",
		 "gif", "image/gif",
		 "jpg", "imgae/jpeg",
		 "xbm", "imgae/x-xbitmap",
		 "mid", "audio/midi"
);


# $BDj?t$NDj5A(B
require 'sys/socket.ph';	# PF_INET, SOCK_STREAM
require 'sys/wait.ph';		# WNOHANG
require 'sys/errno.ph';		# EINTR

# $B%3%^%s%I%i%$%s$N2r@O(B
while($_ = shift(@ARGV)) {
    $svport = shift(@ARGV) if($_ eq '-p');
    $verbose++ if($_ eq '-v');
}

# $B%=%1%C%H$N=`Hw(B
$proto = (getprotobyname('tcp'))[2] || die "getprotobyname";
# $svport = (getservbyname('http', 'tcp'))[2] || die "getservbyname";
$svport = 8080 unless($svport);
$svaddr = pack('S n a4 x8', &AF_INET, $svport, "\0\0\0\0");

# $B%=%1%C%H$N%*!<%W%s(B
socket(SOC, &PF_INET, &SOCK_STREAM, $proto) || die "socket";
bind(SOC, $svaddr) || die "bind: $!";
listen(SOC, 5) || die "listen";

# $B%7%0%J%k$N@_Dj(B
$SIG{'CLD'} = 'sigchld_hdl';	# $B8eJ}8_49@-(B
$SIG{'CHLD'} = 'sigchld_hdl';
$SIG{'INT'} = 'sigint_hdl';

#======================================================================
# $B%a%$%s%k!<%W$H%7%0%J%k%O%s%I%i(B
for(;;) {
    if(!accept(FD, SOC)) {
	# accept()$BCf$K(BSIGCHLD$B$r<u$1$k$3$H$,$"$k(B
	redo if($! == &EINTR);
	die "accept: $!";
    }
    select(FD); $| = 1; select(STDOUT);

    # $B%/%i%$%"%s%H$N%"%I%l%9$rD4$Y$k(B
    $claddr = getpeername(FD);
    ($_, $clport, $clip) = unpack('S n a4', $claddr);
    $cldot = join('.', unpack('C4', $clip));
    $clname = gethostbyaddr($clip, &AF_INET);
    print STDERR "httpd: accepted: $clname($cldot) port=$clport\n";

    # fork$B$7$F%j%/%(%9%H$KEz$($k(B
    die "fork" if(($pid = fork()) < 0);
    if($pid == 0) {
	# $B;R%W%m%;%9(B
	close(SOC);
	$SIG{'CLD'} = 'IGNORE';
	$SIG{'CHLD'} = 'IGNORE';
	&do_httpd;

	print STDERR "child[$$]: service terminated\n";
	exit(0);
    }
    # $B?F%W%m%;%9(B
    close(FD);
}


sub sigint_hdl {
    close(FD);
    close(SOC);
    print STDERR "httpd: SIGINT catched, exitting...\n";
    exit(0);
}


sub sigchld_hdl {
    while(waitpid(-1, &WNOHANG) > 0) {
	# $B2?$b$;$:!$$?$@;R$r@.J)$5$;$F$d$k(B
	print STDERR "httpd: waitpid() done\n";
    }
    $SIG{'CLD'} = 'sigchld_hdl';	# $B8eJ}8_49@-(B
    $SIG{'CHLD'} = 'sigchld_hdl';
}


#======================================================================

sub do_httpd {
    # $B%j%/%(%9%H$r2r@O(B
    $request = <FD>;
    $request =~ s/[\n\r]*$//;
    if($request =~ m/([A-Z]+) (.*) HTTP\/([.0-9])/) {
	($http_method, $http_uri, $http_version) = ($1, $2, $3);
	&read_reqhdr(*http_hdr);
    } elsif($request =~ m/([A-Z]+) (.*)/) {
	# HTTP$B%P!<%8%g%s(B0.9$B$H$_$J$9(B
	($http_method, $http_uri, $http_version) = ($1, $2, '0.9');
    } else {
	# $B%j%/%(%9%H$N7A$,0-$$(B
	print FD "<h1>400 Bad request</h1>\n";
	print STDERR "child[$$]: bad request: $request\n";
	exit(1);
    }
    # URI$B$r%Q%9$H%/%(%j!<J8;zNs$KJ,2r(B
    if($http_uri =~ m/^(.*)\?(.*)$/) {
	($http_path, $http_query) = ($1, $2);
    } else {
	($http_path, $http_query) = ($http_uri, "");
    }

    if($http_method ne "GET"
       && $http_method ne "HEAD"
       && $http_method ne "POST") {
	# $B%5%]!<%H$7$F$$$J$$%a%=%C%I(B
	&error_reply(501, "Not implemented", 
		     "not implemented: $request", "", *ohdr);
    }

    # $BE>Aw$9$Y$-%U%!%$%k$X$N%Q%9$KJQ49(B
    $path = &uri2path($http_path);
    print STDERR "child[$$]: request $http_method $http_uri\n";
    print STDERR "child[$$]: uri2path: $path\n";

    $reply_hdr{"Date"} = &rfc1123(time);
    $reply_hdr{"Server"} = $conf_server;
    if($path =~ m/\.cgi$/ && -x $path) {
	&exec_cgi($path, $http_query);
    } elsif(-f $path && -r $path) {
	&xfer_file($path);
    } else {
	# $B%U%!%$%k$,B8:_$7$J$$$+!$%"%/%;%9ITG=(B
	&error_reply(403, "Access forbidden", 
		     "access forbidden: $request", "", *reply_hdr);
    }
}


sub uri2path {
    local($uri) = @_;
    local($path, @component);

    $uri = $uri . $conf_defdoc if($uri =~ m/\/$/);
    @component = split('/', $uri);
    if($component[1] =~ m/^\~(.*)/) {
	# $BNc$($P(B /~foo/ $B$r(B ~foo/.public_html/index.html$B$KJQ49(B
	$component[1] = (getpwnam($1))[7] . "/" . $conf_usrdir;
	$path = join('/', @component);
    } else {
	$path = $conf_docroot . join('/', @component);
    }
    # $B@hF,$N%@%V$C$?%9%i%C%7%e$r=|5n(B
    $path =~ s(^//)(/);
    $path;
}


sub read_reqhdr {
    # $B%U%#!<%k%I$,(B1$B9T$K<}$^$i$J$$>l9g$KBP1~$7$F$J$$(B
    local(*hdr) = @_;		# %hdr
    local($name, $val);

    while($_ = <FD>) {
	s/[\r\n]*$//;
	last if($_ eq '');
	if(m/^([-\w]+):\s+(.*)/) {
	    ($name, $val) = ($1, $2);
	    print STDERR "child[$$]: request header $name: $val\n";
	    $name =~ tr/A-Z/a-z/;
	    $hdr{$name} = $val;
	}
    }
}


sub write_rephdr {
    # $B%j%W%i%$%X%C%@$rJV$9!%%P!<%8%g%s(B1.0$B0J>e$N(BHTTP$B$G$J$1$l$P;HMQ$7$J$$$3$H!%(B
    local($status, $reason, *outhdr) = @_; # %outhdr

    print FD "HTTP/1.0 $status $reason\n";
    for $_ (keys(%outhdr)) {
	print STDERR "child[$$]: reply header $_: $outhdr{$_}\n";
	print FD "$_: $outhdr{$_}\n";
    }
    print FD "\n";
}

sub rfc1123 {
    # RFC1123$B7A<0$GF|;~$rI=<((B
    local($sec,$min,$hour,$mday,$mon,$year,$wday);
    local(@dayname, @monname);
    @dayname = ("Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat");
    @monname = ("Jan", "Feb", "Mar", "Apr", "May", "Jun",
		"Jul", "Aug", "Sep", "Oct", "Nov", "Dec");

    ($sec,$min,$hour,$mday,$mon,$year,$wday) = gmtime($_[0]);
    sprintf("%s, %02d %s %d %02d:%02d:%02d GMT",
	    $dayname[$wday], $mday, $monname[$mon], $year + 1900,
	    $hour, $min, $sec);
}

sub error_reply {
    local($status, $reason, $explain, $log, *reply_hdr) = @_;

    if($explain eq "") {
	$explain = "<h1>$reason</h1>\n";
    }
    if($http_version > 0.9) {
	&write_rephdr($status, $reason, *reply_hdr);
    }
    print FD $explain;
    print STDERR "child[$$]: $log\n" if($log ne '');
    exit(1);
}

#======================================================================

sub xfer_file {
    local($path) = @_;
    local($len, $buf, $ext, $type, @stat);

    # $B3HD%;R$+$i(BMIME$B%?%$%W$r7hDj$9$k!%(B
    $type = $conf_deftype;
    if($path =~ /\.([^.]+)$/) {
	$ext = $1;
	$type = $conf_exttype{$ext} if($conf_exttype{$ext} ne '');
	$reply_hdr{'Content-type'} = $type;
    }

    # $B%j%W%i%$%X%C%@$NDI2C(B
    if(@stat = stat($path)) {
	$reply_hdr{"Content-length"} = $stat[7];
	$reply_hdr{"Last-modified"} = &rfc1123($stat[9]);
    }

    # $B<B:]$NE>Aw(B
    if(open(TMP, $path)) {
	if($http_version > 0.9) {
	    &write_rephdr(200, "All is right with the world", *reply_hdr);
	}
	while(($len = sysread(TMP, $buf, 65536)) > 0) {
	    syswrite(FD, $buf, $len);
	}
	close(TMP);
    }
}

sub exec_cgi {
    local($path) = @_;

    %ENV = ();			# $B$$$C$?$s$9$Y$F%/%j%"$9$k(B
    $ENV{"PATH"} = "/usr/bin:/usr/ucb";	# $B%;%-%e%j%F%#BP:v(B

    $ENV{"SERVER_SOFTWARE"} = $conf_server;
    $ENV{"SERVER_NAME"} = `hostname`; # $B<jH4$-(B (^^;
    $ENV{"GATEWAY_INTERFACE"} = "CGI/1.1"; # $B$N$D$b$j(B
    $ENV{"SERVER_PROTOCOL"} = "HTTP/1.0"; # $B$K7h$aBG$A(B
    $ENV{"SERVER_PORT"} = $svport;

    $ENV{"REQUEST_METHOD"} = $http_method;
    $ENV{"PATH_INFO"} = ""; # $BL$<BAu(B
    $ENV{"PATH_TRANSLATED"} = ""; # $BL$<BAu(B
    $ENV{"SCRIPT_NAME"} = $http_path;
    $ENV{"QUERY_STRING"} = $http_query if($http_query);
    $ENV{"REMOTE_HOST"} = $clname;
    $ENV{"REMOTE_ADDR"} = $cldot;
    # $BG'>Z5!9=$OFq$7$$$N$GL$<BAu(B
    # $ENV{"AUTH_TYPE", "REMOTE_USER", "REMOTE_IDENT"} = ("", "", "");

    # $B$3$l$i$O%j%/%(%9%H%X%C%@$+$i>pJs$rF@$k(B
    $ENV{"CONTENT_LENGTH"} = $http_hdr{"content-length"} if($http_hdr{"content-length"});
    $ENV{"CONTENT_TYPE"} = $http_hdr{"content-type"} if($http_hdr{"content-type"});
    $ENV{"HTTP_ACCEPT"} = $http_hdr{"accept"} if($http_hdr{"accept"});
    $ENV{"HTTP_USER_AGENT"} = $http_hdr{"user_agent"} if($http_hdr{"user_agent"});
    $ENV{"HTTP_REFERER"} = $http_hdr{"referer"} if($http_hdr{"referer"});
 
    # $B%j%W%i%$%X%C%@$r@8@.(B
    if($http_version > 0.9) {
	print FD "HTTP/1.0 200 Ok\n";
	print FD "Date: " . &rfc1123(time) . "\n";
	print FD "Server: $conf_server\n";
    }

    # $B%U%!%$%k%G%#%9%/%j%W%?$rJ#@=$7$F!$I8=`=PNO$r%=%1%C%H$K$9$k(B
    close(STDOUT); open(STDOUT, ">&FD");
    close(STDIN);
    select(STDOUT); $| = 1;

    # $B<B9T(B
    if($http_method eq "POST") {
	# $B%=%1%C%H$+$i%j%/%(%9%HJ8;zNs$rFI$s$G$*$/(B
	$buf = <FD>;
	close(FD);

	open(TMP, "|$path");
	print TMP $buf;
	close(TMP);
	exit(0);
    } else {
	close(FD);
	exec($path);
    }
}



# TODO:
# /foo/../bar/$B$r5v$7$F$7$^$&(B
# $BJ#?t$N(BAccept: $B%U%#!<%k%I$KBP1~$7$F$$$J$$(B
# POST$B;~$K!$$b$7$+$9$k$H$9$Y$F$N%j%/%(%9%HJ8;zNs$rFI$^$J$$$+$b$7$l$J$$(B
